from sys import modules
from raser.geometry import R3dDetector
from raser.pyfenics import FenicsCal
from raser.g4particles import Particles
from raser.calcurrent import CalCurrent
from raser.elereadout import Amplifier
from raser.setting import Setting

from raser.geometry import R2dDetector
from raser.pyfenics import FenicsCal2D
from raser.calcurrent import CalCurrent2D
from raser.model import Material
from raser.model import Avalanche
from raser.source import MIPs
